import sys
from pytorch_lightning.callbacks import TQDMProgressBar


class FixedProgressBar(TQDMProgressBar):
    """
    A custom progress bar for PyTorch Lightning which addresses potential compatibility issues with some IDEs that
    do not display the default progress bar properly.

    This progress bar inherits from the default TQDMProgressBar provided by PyTorch Lightning and modifies the
    initialization of the progress bars for validation, prediction, and testing.

    Example Usage
    --------------
    ```python

    callbacks = []
    callbacks.append(FixedProgressBar())

    # add any other callbacks if needed

    trainer = pl.Trainer(callbacks=callbacks,
                         **params)
    --------------
    """
    def init_validation_tqdm(self):
        """
        Initialize the progress bar for validation phase.

        Returns
        ----------
        bar : tqdm.tqdm
            The tqdm progress bar object for validation.
        """
        bar = super().init_validation_tqdm()

        # Disable the progress bar if stdout is not a TTY (i.e., it's not an interactive terminal)
        if not sys.stdout.isatty():
            bar.disable = True
        return bar

    def init_predict_tqdm(self):
        """
        Initialize the progress bar for prediction phase.

        Returns
        ----------
        bar : tqdm.tqdm
            The tqdm progress bar object for predicting.
        """
        bar = super().init_predict_tqdm()

        # Disable the progress bar if stdout is not a TTY (i.e., it's not an interactive terminal)
        if not sys.stdout.isatty():
            bar.disable = True
        return bar

    def init_test_tqdm(self):
        """
        Initialize the progress bar for testing phase.

        Returns
        ----------
        bar : tqdm.tqdm
            TThe tqdm progress bar object for testing.
        """
        bar = super().init_test_tqdm()

        # Disable the progress bar if stdout is not a TTY (i.e., it's not an interactive terminal)
        if not sys.stdout.isatty():
            bar.disable = True
        return bar
