import torch.nn as nn
from ..data.SpecialTokens import SpecialTokens

PAD_TOKEN, START_TOKEN, END_TOKEN = SpecialTokens.PAD_TOKEN, SpecialTokens.START_TOKEN, SpecialTokens.END_TOKEN


class VanillaClassifier(nn.Module):
    """
    A customizable classifier that supports Linear, BiLSTM, or Transformer layers to map input features to a fixed
    number of labels.

    Attributes:
    ----------
    tag_mapper : torch.nn.Module, optional
        Layer for mapping to the tag space. Can be LSTM, Transformer, or None (for a linear mapping).
    linear : torch.nn.Linear
        Linear layer mapping the input (or the output of tag_mapper) to the label space.

    """
    def __init__(self,
                 label_to_id_map: dict,
                 input_dim: int,
                 to_tag_space: str = "linear"):
        """
        Parameters:
        ----------
        label_to_id_map : dict
            Mapping of labels to their respective unique identifiers
        input_dim : int
            Dimensionality of the input features.
        to_tag_space : str, optional (default: "linear")
            Specifies the tag space mapping, either "linear", "bilstm", or "transformer".

        """
        super(VanillaClassifier, self).__init__()
        num_labels = len(label_to_id_map) - len({START_TOKEN, END_TOKEN, PAD_TOKEN} & set(label_to_id_map.keys()))

        if to_tag_space == "bilstm":
            self.tag_mapper = nn.LSTM(input_dim, input_dim, bidirectional=True, batch_first=True)
            self.linear = nn.Linear(in_features=input_dim * 2, out_features=num_labels)
        elif to_tag_space == "transformer":
            self.tag_mapper = nn.Transformer(input_dim, input_dim)
            self.linear = nn.Linear(in_features=input_dim, out_features=num_labels)
        else:  # "linear"
            self.tag_mapper = None
            self.linear = nn.Linear(in_features=input_dim, out_features=num_labels)

    def forward(self, x):
        """
        Parameters:
        ----------
        x : torch.Tensor
            Input tensor with features which should output of encoder on the word level.
        """
        if self.tag_mapper:
            x, _ = self.tag_mapper(x)
        out = self.linear(x)
        return out

