from torch.utils.data import Dataset
from pathlib import Path
from .SpecialTokens import SpecialTokens
from .NerTokenizer import NerTokenizer
from typing import Dict

CHAR_PAD_TOKEN, CHAR_START_TOKEN, CHAR_END_TOKEN, CHAR_UNKNOWN_TOKEN, PAD_TOKEN, START_TOKEN, END_TOKEN = \
    SpecialTokens.CHAR_PAD_TOKEN, SpecialTokens.CHAR_START_TOKEN, SpecialTokens.CHAR_END_TOKEN, \
    SpecialTokens.CHAR_UNKNOWN_TOKEN, SpecialTokens.PAD_TOKEN, SpecialTokens.START_TOKEN, SpecialTokens.END_TOKEN


class NerDataset(Dataset):
    """
    Named Entity Recognition (NER) Dataset.
    A PyTorch dataset class for reading and tokenizing NER data from a file. The file with data should adhere to the
    following structure:
    - Each line consists of an <entity> <tag> pair, delimited by a space.
    - Consecutive lines represent words in the same sentence, in their original order.
    - The start of a new sentence is indicated by a metadata line: `# sent_id = <n>`, where `<n>` is a unique
      sentence identifier within the file.
    - A newline character demarcates the end of each sentence.

    Attributes:
    ----------
    tokenizer : NerTokenizer
        Custom NerTokenizer object for encoding the entities.
    sentences : List[List]
        List of list of entities - strings. Inner list can be understood as sentence.
    sentence_id : List[int]
        List of unique sentence IDs.
    labels : List[List]
        List of lists of labels corresponding to each tokenized sentence.
    max_sentence_len : int
        Length of the longest sentence in the dataset.
    max_word_len : int
        Length of the longest word in the dataset.

    """

    def __init__(self,
                 tokenizer: NerTokenizer,
                 file_path: Path,
                 encoding: str = 'utf-8') -> None:
        """
        Initializes the dataset with a tokenizer, reads the data, and processes it. Tokenizer is expected to be a custom
        NerTokenizer. Other tokenizers were not tested.

        Parameters:
        ----------
        tokenizer : NerTokenizer
            The tokenizer for encoding entities and labels.
        file_path : Path
            File path to the data.
        encoding : str (optional)
            Encoding of the data file. Defaults to 'utf-8'.
        """
        # save tokenizer
        self.tokenizer = tokenizer

        # data holders
        self.sentences = []
        self.sentence_id = []
        self.labels = []

        # additional information about data
        self.max_sentence_len = 0
        self.max_word_len = 0

        # read data
        self._read_file(file_path=file_path,
                        encoding=encoding)

    def __len__(self) -> int:
        return len(self.sentences)

    def __getitem__(self,
                    index: int) -> Dict:
        # retrieve data
        words = self.sentences[index]
        labels = self.labels[index]

        # use tokenizer to encode it
        output = self.tokenizer.encode(input_entities=words,
                                       labels=labels)
        return output

    def _read_file(self,
                   file_path: Path,
                   encoding: str) -> None:
        """
        Reads the data file and processes it, populating sentences, labels, and other metadata.

        Parameters:
        ----------
        file_path : Path
            Path to the data file.
        encoding : str
            Encoding of the data file.
        """
        # Reading file
        with open(file_path, 'r', encoding=encoding) as f:
            words = []
            labels = []

            for line in f:
                # at the end of each sequence of entities from the same sentence there should be "\n"
                if line == '\n':
                    # update longest sentence
                    self.max_sentence_len = max(self.max_sentence_len, len(words))

                    # save labels and words
                    self.sentences.append(words)
                    self.labels.append(labels)
                    words = []
                    labels = []

                # save sentence id at the beginning of every sequence of entities
                elif line.startswith("# sent_id ="):
                    self.sentence_id.append(line.split()[3])

                # otherwise it should be entity and tag separated by space
                else:
                    parts = line.strip().split()
                    words.append(parts[0])
                    labels.append(parts[-1])

                    # update longest word length
                    self.max_word_len = max(self.max_word_len, len(parts[0]))

        self.max_word_len += 2  # for <b> and <e> tokens
