from collections import Counter
from pathlib import Path
from typing import Dict
from .SpecialTokens import SpecialTokens

CHAR_PAD_TOKEN, CHAR_START_TOKEN, CHAR_END_TOKEN, CHAR_UNKNOWN_TOKEN, PAD_TOKEN, START_TOKEN, END_TOKEN = \
    SpecialTokens.CHAR_PAD_TOKEN, SpecialTokens.CHAR_START_TOKEN, SpecialTokens.CHAR_END_TOKEN, \
    SpecialTokens.CHAR_UNKNOWN_TOKEN, SpecialTokens.PAD_TOKEN, SpecialTokens.START_TOKEN, SpecialTokens.END_TOKEN


def create_tag2id(file_path: Path,
                  encoding: str = 'utf-8',
                  include_special_tokens: bool = False) -> Dict[str, int]:
    """
    Generate a unique ID mapping for tags sourced from a specified file. The file should adhere to the following
    structure:
    1. Each line consists of an <entity> <tag> pair, delimited by a space.
    2. Consecutive lines represent words in the same sentence, in their original order.
    3. The start of a new sentence is indicated by a metadata line: `# sent_id = <n>`, where `<n>` is a unique
      sentence identifier within the file.
    4. A newline character demarcates the end of each sentence.

   Parameters:
    ----------
    file_path : Path
        Path to the file containing words and tags.
    encoding : str, optional
        Encoding of the file. Defaults to 'utf-8'.
    include_special_tokens : bool, optional
        If true, special tokens such as START and END are added to the mapping. Defaults to False.

    Returns:
    -------
    tag2id : Dict[str, int]
        A dictionary mapping tags to unique IDs.
    """
    tag2id = set()

    # Reading the file
    with open(file_path, 'r', encoding=encoding) as f:
        for line in f:
            # Omit spaces between sentences and sentence ID annotations
            if line == '\n' or line.startswith("# sent_id ="):
                continue
            # Otherwise, the line should be a word-entity pair separated by space
            else:
                tag = line.strip().split(" ")[-1]
                if tag != "O":
                    # Add both beginning and inside tags for entities
                    assert len(tag.split("-")) == 2, "Tags should be in format <entity>-<tag>, too many dashes detected"
                    tag2id.add("B-"+tag.split("-")[1])
                    tag2id.add("I-" + tag.split("-")[1])
                else:
                    tag2id.add(tag)

    # Convert the set to a dictionary with sorted tags mapped to unique IDs
    tag2id = {tag: i for i, tag in enumerate(sorted(tag2id))}
    tag2id[PAD_TOKEN] = len(tag2id)
    if include_special_tokens:
        tag2id[START_TOKEN] = len(tag2id)
        tag2id[END_TOKEN] = len(tag2id)

    return tag2id


def create_char2id(file_path: Path,
                   encoding: str = "utf-8",
                   discard_least_common: float = 0.0005) -> Dict[str, int]:
    """
    Create a mapping from characters in the file to unique IDs, potentially discarding the least common characters. File
    format is expected to be the same as indicated in create_tag2id() documentation

    Parameters:
    ----------
    file_path : Path
        Path to the file containing words and tags.
    encoding : str, optional
        Encoding of the file. Defaults to 'utf-8'.
    discard_least_common : float, optional
        Proportion of total characters to discard from the mapping based on their frequency. Defaults to 0.0005.

    Returns:
    -------
    char2id : Dict[str, int]
        A dictionary mapping characters to unique IDs.
    """

    char2id = {CHAR_PAD_TOKEN, CHAR_START_TOKEN, CHAR_END_TOKEN, CHAR_UNKNOWN_TOKEN}
    char_counter = Counter()

    # Reading the file
    with open(file_path, 'r', encoding=encoding) as f:
        for line in f:
            # Omit spaces between sentences and sentence ID annotations
            if line == '\n' or line.startswith("# sent_id ="):
                continue
            # Otherwise, the line should be a word-entity pair separated by space
            else:
                word = line.strip().split()[0]
                char_counter.update([char for char in word])

    total_chars = sum(char_counter.values())
    cutoff = total_chars * discard_least_common  # Calculate the cutoff frequency

    sorted_chars = sorted(char_counter.items(), key=lambda x: x[1])  # Sort characters by frequency

    # Discard the least common characters based on the cutoff
    current_count = 0
    for char, count in sorted_chars:
        current_count += count
        if current_count > cutoff:
            char2id.add(char)

    # Convert the set to a dictionary mapped to unique IDs
    char2id = {char: i for i, char in enumerate(char2id)}
    return char2id


def calculate_longest_word(file_path: Path,
                           encoding: str = "utf-8") -> int:
    """Extract maximum character length of a word from a file."""
    max_length = 0
    # Reading the file
    with open(file_path, 'r', encoding=encoding) as f:
        for line in f:
            # Omit spaces between sentences and sentence ID annotations
            if line == '\n' or line.startswith("# sent_id ="):
                continue
            # Otherwise, the line should be a word-entity pair separated by space
            else:
                word = line.strip().split()[0]
                max_length = max(max_length, len(word))

    return max_length
