from .TokenPooling import TokenPooling
import torch.nn as nn
import torch
from transformers import AutoModel, BertModel


class BertEmbedder(nn.Module):
    """
    Embeds text using a BERT model and provides additional options for token pooling and projections.

    This class wraps either an AutoModel or BertModel from the Hugging Face transformers library, allowing optional
    projection and pooling layers on top of the BERT embeddings. The pooling strategy and other options can be
    customized based on the parameters provided.

    Attributes:
    ----------
    bert : transformers.AutoModel or transformers.BertModel
        The BERT model from transformers that's either an AutoModel or BertModel, depending on the specified
        pretrained_model_type.
    projection_layer : torch.nn.Linear, optional
        A linear layer to project the BERT embeddings to a different dimension. Only initialized if projection_dimension
        is provided.
    token_pooling : TokenPooling, optional
        A token pooling layer to pool tokens based on the specified strategy. Only initialized if token_pooling is set
        to True.
    output_dimension : int
        The dimension of the output embeddings, determined by either the BERT hidden size or the projection_dimension
        if specified.

    """
    def __init__(self,
                 pretrained_model_name: str,
                 pretrained_model_type: str,
                 projection_dimension: int = None,
                 freeze_bert: bool = True,
                 token_pooling: bool = True,
                 use_start_end_token: bool = False,
                 pooling_strategy: str = "max") -> None:
        """
         Parameters:
        ----------
        pretrained_model_name : str
            The name of the pre-trained BERT model to be loaded (e.g., 'bert-base-uncased').
        pretrained_model_type : str
            The type of BERT model to be used ('AutoModel' or 'BertModel').
        projection_dimension : int, optional
            The dimension for the projection layer applied to BERT embeddings. If specified, a linear layer
            is initialized for the projection. (default: None)
        freeze_bert : bool, optional
            Whether to freeze the parameters of the BERT model, disabling further training. (default: True)
        token_pooling : bool, optional
            Whether to initialize a token pooling layer based on the specified strategy. (default: True)
        use_start_end_token : bool, optional
            Whether to use start and end tokens in token pooling. Only relevant if token_pooling is True.(default: False)
        pooling_strategy : str, optional
            The strategy to be used for token pooling if token_pooling is set to True. (default: "max")
        """
        assert pretrained_model_type in ['AutoModel', 'BertModel'], \
            'pretrained_model_type must be either BertModel or AutoModel'

        super(BertEmbedder, self).__init__()

        # initialize bert model
        if pretrained_model_type == 'AutoModel':
            self.bert = AutoModel.from_pretrained(pretrained_model_name)
        elif pretrained_model_type == 'BertModel':
            self.bert = BertModel.from_pretrained(pretrained_model_name)

        # freeze bert model
        if freeze_bert:
            for param in self.bert.parameters():
                param.requires_grad = False

        # initialize projection layer if projection dimension is given
        if projection_dimension:
            self.projection_layer = nn.Linear(in_features=self.bert.config.hidden_size,
                                              out_features=projection_dimension)
            self.output_dimension = projection_dimension
        else:
            self.output_dimension = self.bert.config.hidden_size
            self.projection_layer = None

        # initialize token pooling layer if specified
        if token_pooling:
            self.token_pooling = TokenPooling(pooling_strategy=pooling_strategy,
                                              use_start_end_token=use_start_end_token)
        else:
            self.token_pooling = None

    def forward(self,
                input_ids: torch.Tensor,
                attention_mask: torch.Tensor,
                word_ids: torch.Tensor = None) -> torch.Tensor:
        """
        Performs the forward pass on the given inputs and returns the corresponding BERT embeddings, with optional
        token pooling and projection
        """
        outputs = self.bert(input_ids,
                            attention_mask=attention_mask)
        outputs = outputs.last_hidden_state

        # apply token pooling if specified
        if self.token_pooling is not None:
            outputs = self.token_pooling(bert_outputs=outputs,
                                         word_ids=word_ids)

        # apply projection layer if specified
        if self.projection_layer is not None:
            outputs = self.projection_layer(outputs)

        return outputs
