from ..data.SpecialTokens import SpecialTokens
from seqeval.metrics import f1_score, recall_score, precision_score, classification_report


START_TOKEN, END_TOKEN, PAD_TOKEN = SpecialTokens.START_TOKEN, SpecialTokens.END_TOKEN, SpecialTokens.PAD_TOKEN


class MetricHolder:
    """
    A utility class to hold and manage metrics related to sequence labeling tasks. Built in methods are provided to
    print recall, F1 and precision scores as well as classification report.

    Attributes:
    ----------
    targets : list
        A list to store the true labels of sequences.
    predictions : list
        A list to store the predicted labels of sequences.
    label_to_idx : dict
        A dictionary to map label names to their respective integer indices.
    idx_to_label : dict
        A dictionary to map integer indices to their respective label names.
    ignore_idx : int
        Index of the token to be ignored during metric calculation. Default is -100.
    """
    def __init__(self,
                 label_to_idx: dict = None,
                 idx_to_label: dict = None,
                 ignore_idx: int = -100):
        """
        Initialize a MetricHolder instance.

        Parameters:
        ----------
        label_to_idx : dict
            A dictionary to map label names to their respective integer indices.
        idx_to_label : dict
            A dictionary to map integer indices to their respective label names.
        ignore_idx : int
            Index of the token to be ignored during metric calculation. Default is -100. This should be the index of
            tokens like start, end and padding tokens.
        """
        self.targets = []
        self.predictions = []

        assert label_to_idx or idx_to_label, "Specify label_to_idx or idx_to_label"
        if label_to_idx:
            self.label_to_idx = label_to_idx
            self.idx_to_label = {idx: label for label, idx in label_to_idx.items()}
        else:
            self.idx_to_label = idx_to_label
            self.label_to_idx = {label: idx for idx, label in idx_to_label.items()}

        self.ignore_idx = ignore_idx

    def update(self, predictions, targets) -> None:
        """
        Update the internal targets and predictions with new data.

        Parameters:
        ----------
            predictions (list): A list of predicted labels for sequences.
            targets (list): A list of true labels for sequences.
        """
        # flatten predictions and targets
        predictions = predictions.tolist()
        targets = targets.tolist()

        # Delete padded values
        predictions = [[y_hat for y, y_hat in zip(y_sentence,y_sentence_hat) if y != self.ignore_idx] for y_sentence, y_sentence_hat in zip(targets, predictions)]
        targets = [[y for y in y_sentence if y != self.ignore_idx] for y_sentence in targets]

        # Translate to labels
        predictions = [[self.idx_to_label[y_hat] for y_hat in y_sentence_hat] for y_sentence_hat in predictions]
        # Fixing issue with some predictions being end or pad tag to by default be O tag
        predictions = [[y_hat if y_hat not in [END_TOKEN, PAD_TOKEN, START_TOKEN] else "O" for y_hat in y_sentence_hat]
                       for y_sentence_hat in predictions]
        targets = [[self.idx_to_label[y] for y in y_sentence] for y_sentence in targets]

        # Update targets and predictions
        self.targets += targets
        self.predictions += predictions

    def reset(self) -> None:
        """
        Reset the internal targets and predictions to empty lists.
        """
        self.targets = []
        self.predictions = []

    def get_classification_report(self) -> str:
        """
        Retrieve a detailed classification report for the current targets and predictions.

        Returns:
        ----------
            str: The classification report in tabular format in string.
        """
        report = classification_report(self.targets,
                                       self.predictions,
                                       zero_division=1)
        return report

    def get_metrics(self,
                    stage: str = None) -> dict:
        """
        Calculate and retrieve key metrics for the current targets and predictions.

        Parameters:
        ----------
        stage : str
            The name of the current stage (e.g., 'train', 'val', 'test') to be used as a prefix for metric keys.

        Returns:
        ----------
        dict : A dictionary containing the precision, recall, and F1-score.
        """
        assert stage is not None, "Specify stage"

        return {
            f'{stage}_precision': precision_score(self.targets, self.predictions, average="micro", zero_division=1),
            f'{stage}_recall': recall_score(self.targets, self.predictions, average="micro", zero_division=1),
            f'{stage}_f1': f1_score(self.targets, self.predictions, average="micro", zero_division=1)
        }
