from string import Template
from pathlib import Path
from utils import (
    download_file, 
    extract_tgz, 
    build_list_of_uds_to_train, 
    download_fasttext,
    write_to_csv,
    train_one_model)
import pandas as pd
import argparse

# Argument parsing
parser = argparse.ArgumentParser()
parser.add_argument("--treebank_id", type=str, default="", help="ID of the treebank")
parser.add_argument("--treebank_version", type=str, default="", help="Version of the treebank")
parser.add_argument("--output_dir", type=str, default="", help="Base directory for output")
args = parser.parse_args()

url = Template("https://lindat.mff.cuni.cz/repository/xmlui/bitstream/handle/11234/$treebank_id/ud-treebanks-v$version.tgz")

def main():
    treebank_id = args.treebank_id
    treebank_version = args.treebank_version
    output_dir = args.output_dir
    ud_output_dir = str(Path(output_dir) / f'ud-treebanks-v{treebank_version}.tgz')

    # make output directory if it doesn't exist
    Path(output_dir).mkdir(parents=True, exist_ok=True)
    (Path(output_dir) / "results").mkdir(parents=True, exist_ok=True)

    ud_url = url.substitute(treebank_id=treebank_id,
                            version=treebank_version)
    # Download UD treebank
    if not Path(ud_output_dir).exists():
        print("Downloading UD treebank")
        download_file(url=ud_url,
                      local_filename=ud_output_dir)
    else:
        print("UD treebank already downloaded")

    # Extract UD treebank    
    if not (Path(output_dir) / f'ud-treebanks-v{treebank_version}').exists():
        print("Extracting UD treebank")
        extract_tgz(tgz_path=ud_output_dir,
                    extract_path=output_dir)
    else :
        print("UD treebank already extracted")

    print("Checking UD data")
    ud_folder_directory = output_dir + f'/ud-treebanks-v{treebank_version}'
    training_queue = build_list_of_uds_to_train(UD_dir=ud_folder_directory)

    # Read UDs with completed training 
    completed_training_path = Path(output_dir) / "results" / "completed_training.txt"
    completed_training_path.touch(exist_ok=True) 
    with open(completed_training_path, 'r', encoding='utf-8') as file:
        completed_training = [line.strip() for line in file]

    # Train on transformer models
    for ud_treebank in training_queue.keys():
        if ud_treebank not in completed_training:
            print(f"Training {ud_treebank}")
            try:
                # Start training 
                train_one_model(treebank = ud_treebank,
                                treebanks_dir = ud_folder_directory,
                                output_dir = output_dir,
                                targets = training_queue[ud_treebank],
                                use_fasttext = False)
                
                # Save info that training was completed
                with open(completed_training_path, 'a') as file: file.write(ud_treebank+"\n")

                # Delete this info from skipped_training.csv if there is any
                df = pd.read_csv(Path(output_dir) / "results" / "skipped_training.csv")
                                 
                filtered_df = df[~df.iloc[:, 0].str.contains(ud_treebank)]
                filtered_df.to_csv(Path(output_dir) / "results" / "skipped_training.csv", 
                                   index=False)
            except Exception as e:
                error = str(e)
                write_to_csv(treebank=ud_treebank,
                             reason=error, 
                             csv_file_path=Path(output_dir) / "results" / "skipped_training.csv")


if __name__ == "__main__":
    main()
