import requests
from tqdm import tqdm
import tarfile
from pathlib import Path
import csv
import shutil
from constants import (
    LINKS, 
    CODE_2_LANG, 
    UD_2_BATCH_SIZE,
    LANG2TRANSFORMER,
    LANG_UD2FASTTEXT)
import subprocess
import os


def download_file(url : str,
                  local_filename : str,
                  show_progress : bool=True):
    """
    Downloads a file from the given URL and saves it locally.

    Args:
        url (str): The URL of the file to download.
        local_filename (str): The path and filename to save the downloaded file locally.
        show_progress (bool, optional): Whether to show a progress bar during the download. 
            Defaults to True.
    """
    # Send a GET request to the URL with stream=True to stream the download
    with requests.get(url, stream=True) as response:
        # Raise an error for bad responses
        response.raise_for_status()

        # Get the total size of the file from the Content-Length header
        total_size_in_bytes = int(response.headers.get('content-length', 0))

        # Open a local file with write-binary mode
        with open(local_filename, 'wb') as file:
            if show_progress:
                # Use tqdm for a progress bar if show_progress is True
                with tqdm(
                        desc=local_filename,
                        total=total_size_in_bytes,
                        unit='iB',
                        unit_scale=True,
                        unit_divisor=1024,
                ) as progress_bar:
                    for chunk in response.iter_content(chunk_size=1024):
                        size = file.write(chunk)
                        progress_bar.update(size)
            else:
                # If not showing progress, just write the file without a progress bar
                for chunk in response.iter_content(chunk_size=1024):
                    file.write(chunk)

    print(f"File downloaded and saved as {local_filename}")


def download_fasttext(output_dir: str):
    """
    Downloads FastText vectors for different languages.

    Args:
        output_dir (str): The directory where the FastText vectors will be downloaded.
    """
    output_dir = Path(output_dir)
    for link in LINKS:
        lang_code = link.split(".")[-4]

        if lang_code not in CODE_2_LANG:
            print(f"Unknown code {lang_code}.")
            continue

        output_file = output_dir / CODE_2_LANG[lang_code]
        output_file.mkdir(exist_ok=True, parents=True)
        if (output_file / 'vectors.vec.gz').exists():
            continue

        download_file(url=link,
                      local_filename=str(output_file / 'vectors.vec.gz'),
                      show_progress=False)


def extract_tgz(tgz_path : str, 
                extract_path : str ="."):
    """
    Extracts the contents of a .tgz file to the specified extract path.

    Args:
        tgz_path (str): The path to the .tgz file.
        extract_path (str, optional): The path where the contents will be extracted. Defaults to the current directory.
    """
    # Open the .tgz file
    with tarfile.open(tgz_path, "r:gz") as tar:
        # Get the total number of items within the archive for the progress bar
        total = len(tar.getmembers())

        # Setup tqdm progress bar
        with tqdm(total=total, unit='files', desc="Extracting") as progress_bar:
            # Extract each member individually
            for member in tar.getmembers():
                tar.extract(member, path=extract_path)
                # Update the progress bar per file extracted
                progress_bar.update(1)
        print(f"Files have been extracted to: {extract_path}")


def write_to_csv(treebank : str, 
                 reason : str,
                 csv_file_path : Path):   
    # Open the CSV file and append the directory name and reason
    with open(csv_file_path, mode='a', newline='') as file:
        writer = csv.writer(file)
        writer.writerow([treebank, reason])


def build_list_of_uds_to_train(UD_dir: str):
    """
    Builds a dictionary of UD treebanks to train with their respective targets training commands. It ommits UDs that
    have training file of size less than 1000 bytes, corrupted files (less than 10 columns) or missing files. 
    All of reasons are saved in skipped_training.csv

    Args:
        UD_dir (str): The directory path containing the UD treebanks.

    Returns:
        dict: A dictionary mapping UD treebank names to their training commands.
    """
    UD_dir = Path(UD_dir)
    csv_file_path = UD_dir.parent / "results" / "skipped_training.csv"
    UD_to_targets = {}

    # Create file with skipped UD treebanks
    with open(csv_file_path, mode='w', newline='') as file:
        writer = csv.writer(file)
        writer.writerow(["Directory Name", "Reason"])

    # Iterate through each item in the directory
    for item in tqdm(UD_dir.iterdir()):
        # Check if the item is a directory
        if item.is_dir():
            
            skipped = False
            # Get files in directory (ignoring directories)
            onlyfiles = [f for f in item.iterdir() if f.is_file()]

            # Check for the presence of specific files
            has_dev = any(["dev" in f.name and f.suffix == ".conllu" for f in onlyfiles])
            has_test = any(["test" in f.name and f.suffix == ".conllu" for f in onlyfiles])
            has_train = any(["train" in f.name and f.suffix == ".conllu" for f in onlyfiles])

            if not (has_train and has_dev and has_test):
                write_to_csv(treebank=item.name,
                             reason="Dev or test or train file missing",
                             csv_file_path=csv_file_path)
                continue
            
            # If training file are too small - skip training
            for f in item.iterdir():
                tmp_path = Path()
                if f.suffix == ".conllu" and ("dev" in f.name or "test" in f.name or "train" in f.name):
                    tmp_path = f
                    if "train" in f.name:
                        train_file = f
                if tmp_path and tmp_path.is_file() and tmp_path.stat().st_size < 1000:
                    write_to_csv(treebank=item.name,
                                 reason="Training file less than 1000 bytes",
                                 csv_file_path=csv_file_path)
                    continue

            lemmas, upos, xpos, feats, head, deprel = [], [], [], [], [], []

            # Checking what data is inside training file
            with train_file.open('r', encoding='utf-8') as rf:
                sentences_processed = 0
                for line in rf:
                    # omit lines that are comments
                    if line.startswith("#"):
                        continue

                    # always read only 10 sentences
                    if line == "\n":
                        if sentences_processed == 10:
                            break
                        else:
                            sentences_processed += 1
                            continue

                    words = line.split("\t")
                    if len(words) == 10:
                        lemmas.append(words[2])
                        upos.append(words[3])
                        xpos.append(words[4])
                        feats.append(words[5])
                        head.append(words[6])
                        deprel.append(words[7])
                    else:
                        write_to_csv(treebank=item.name,
                                     reason="Training file corrupted - number of columns is less than 10.",
                                     csv_file_path=csv_file_path)
                        skipped = True
                        break

            
            if skipped:
                continue
            
            # Create targets command by adding 
            command = " --targets head,deprel"
            if set(lemmas) != "-":
                command += ",lemma"
            if set(upos) != "-":
                command += ",upostag"
            if set(xpos) != "-":
                command += ",xpostag"
            if set(feats) != "-":
                command += ",feats"

            # Save training command
            UD_to_targets[item.name] = command
    
    return UD_to_targets
        

def execute_command(command: str, output_file=None):
    """
    Executes a command in the system shell.

    Args:
        command (str): The command to be executed.
        output_file (str, optional): The file to redirect the command output to. Defaults to None.
    """
    env = os.environ.copy()
    command = [c for c in command.split() if c.strip()]
    if output_file:
        with open(output_file, "w") as f:
            subprocess.run(command, check=True, stdout=f, env=env)
    else:
        subprocess.run(command, check=True, env=env)


def train_one_model(treebank : str,
                    treebanks_dir : str,
                    output_dir : str,
                    targets : str,
                    use_fasttext : bool = True):
    # Create LANG_2_CODE dict
    LANG_2_CODE = {value: key for key, value in CODE_2_LANG.items()}

    treebank_dir = Path(treebanks_dir) / treebank
    treebank_parts = treebank[3:].split("-")
    language = treebank_parts[0]
    if language in LANG_UD2FASTTEXT:
        language = LANG_UD2FASTTEXT[language]

    # Find training and dev files
    files = list(treebank_dir.iterdir())

    training_file = [f for f in files if "train" in f.name and ".conllu" in f.name]
    assert len(training_file) == 1, f"Multiple training file."
    training_file_path = training_file[0]

    valid_file = [f for f in files if "dev" in f.name and ".conllu" in f.name]
    assert len(valid_file) == 1, f"Multiple validation file."
    valid_file_path = valid_file[0]

    # Find fasttexts
    if use_fasttext:
        embeddings_dir = Path(output_dir) / "fasttext" / language
        if not embeddings_dir.exists():
            raise FileNotFoundError(f"Fasttext for {language} does not exist. If you want to use fasttext for different language, specify mapping in LANG_UD2FASTTEXT.")
        embeddings_file = [f for f in embeddings_dir.iterdir() if "vectors" in f.name and ".vec" in f.name]
        assert len(embeddings_file) == 1, f"Couldn't find embeddings file."
        embeddings_file = embeddings_file[0]
    else:
        # Throw error if transformer model is not specified
        if not LANG_2_CODE[language] in LANG2TRANSFORMER.keys():
            raise ValueError(f"Specify transformer model for language code: {LANG_2_CODE[language]}.")

    # Create serialization directory 
    serialization_dir = Path(output_dir) / "results" / "serialization_directories" / treebank
    if serialization_dir.exists() and serialization_dir.is_dir():
        # Remove the directory and all its contents
        shutil.rmtree(serialization_dir)
    serialization_dir.mkdir(exist_ok=True, parents=True)

    # Create command to run training
    training_data_path = f"--training_data_path {training_file_path} "
    validation_data_path = f"--validation_data_path {valid_file_path} "
    if use_fasttext:
        embeddings = f"--pretrained_tokens {embeddings_dir} "
    else:
        transformer_name = LANG2TRANSFORMER[LANG_2_CODE[language]]
        embeddings = f"--pretrained_transformer_name {transformer_name} "
    serialization_dir = f"--serialization_dir {serialization_dir} "
    tokenizer_language = f"--tokenizer_language={language} "

    command = "combo --mode train " + training_data_path + validation_data_path + embeddings + serialization_dir + tokenizer_language
    
    command += targets

    if treebank in UD_2_BATCH_SIZE:
        command += f" --word_batch_size {UD_2_BATCH_SIZE[treebank]}"

    # Start training
    execute_command(command)



